import * as cp from 'child_process';
import * as readline from 'readline';
import * as decoder from 'string_decoder';
import * as protocol from 'typescript/lib/protocol';

export const enum EventTypes {
    ConfigFileDiag = 'configFileDiag',
    RequestCompleted = 'requestCompleted',
    SyntaxDiag = 'syntaxDiag',
    SementicDiag = 'semanticDiag',
}

export const enum CommandTypes {
    Brace = 'brace',
    BraceCompletion = 'braceCompletion',
    Change = 'change',
    Close = 'close',
    Completions = 'completions',
    CompletionDetails = 'completionEntryDetails',
    CompileOnSaveAffectedFileList = 'compileOnSaveAffectedFileList',
    CompileOnSaveEmitFile = 'compileOnSaveEmitFile',
    Configure = 'configure',
    Definition = 'definition',
    Implementation = 'implementation',
    Exit = 'exit',
    Format = 'format',
    Formatonkey = 'formatonkey',
    Geterr = 'geterr',
    GeterrForProject = 'geterrForProject',
    SemanticDiagnosticsSync = 'semanticDiagnosticsSync',
    SyntacticDiagnosticsSync = 'syntacticDiagnosticsSync',
    NavBar = 'navbar',
    Navto = 'navto',
    NavTree = 'navtree',
    NavTreeFull = 'navtree-full',
    Occurrences = 'occurrences',
    DocumentHighlights = 'documentHighlights',
    Open = 'open',
    Quickinfo = 'quickinfo',
    References = 'references',
    Reload = 'reload',
    Rename = 'rename',
    Saveto = 'saveto',
    SignatureHelp = 'signatureHelp',
    TypeDefinition = 'typeDefinition',
    ProjectInfo = 'projectInfo',
    ReloadProjects = 'reloadProjects',
    Unknown = 'unknown',
    OpenExternalProject = 'openExternalProject',
    OpenExternalProjects = 'openExternalProjects',
    CloseExternalProject = 'closeExternalProject',
    TodoComments = 'todoComments',
    Indentation = 'indentation',
    DocCommentTemplate = 'docCommentTemplate',
    CompilerOptionsForInferredProjects = 'compilerOptionsForInferredProjects',
    GetCodeFixes = 'getCodeFixes',
    GetSupportedCodeFixes = 'getSupportedCodeFixes',
    GetApplicableRefactors = 'getApplicableRefactors',
    GetEditsForRefactor = 'getEditsForRefactor',
}

export class Deferred<T> {

    constructor(private operation: string, timeout?: number) {
        setTimeout(() => {
            this.reject(new Error(this.operation + ' timeout'));
        }, timeout || 20000);
    }

    public resolve: (value?: T) => void;
    public reject: (err?: any) => void;

    public promise = new Promise<T>((resolve, reject) => {
        this.resolve = resolve;
        this.reject = reject;
    });
}

export interface TspClientOptions {
    tsserverPath: string;
    onEvent?: (event: protocol.Event) => void;
}

export class TspClient {
    private readlineInterface: readline.ReadLine;
    private tsserverProc: cp.ChildProcess;
    private seq = 0;

    private buffer = '';
    private header: Record<string, any> | undefined;

    private deferreds = {};

    constructor(private options: TspClientOptions) {
    }

    public start() {
        if (this.readlineInterface) {
            return;
        }
        const args: string[] = [];
        console.log(`Starting tsserver : '${this.options.tsserverPath} ${args.join(' ')}'`);
        this.tsserverProc = cp.spawn(this.options.tsserverPath, args);
        this.readlineInterface = readline.createInterface(this.tsserverProc.stdout, this.tsserverProc.stdin, undefined);
        process.on('exit', () => {
            this.readlineInterface.close();
            this.tsserverProc.stdin.destroy();
            this.tsserverProc.kill();
        });
        this.readlineInterface.on('line', line => this.processMessage(line));

        const dec = new decoder.StringDecoder('utf-8');
        this.tsserverProc.stderr.addListener('data', data => {
            const stringMsg = typeof data === 'string' ? data : dec.write(data);
        });
    }

    public notify(command: CommandTypes.Open, args: protocol.OpenRequestArgs);
    public notify(command: CommandTypes.Close, args: protocol.FileRequestArgs);
    public notify(command: CommandTypes.Saveto, args: protocol.SavetoRequestArgs);
    public notify(command: CommandTypes.Change, args: protocol.ChangeRequestArgs);
    public  notify(command: string, args: object) {
        this.sendMessage(command, true, args);
    }

    public request(command: CommandTypes.Definition, args: protocol.FileLocationRequestArgs): Promise<protocol.DefinitionResponse>;
    public request(command: CommandTypes.Format, args: protocol.FormatRequestArgs): Promise<protocol.FormatResponse>;
    public request(command: CommandTypes.GetApplicableRefactors, args: protocol.CodeFixRequestArgs): Promise<protocol.GetCodeFixesResponse>;
    public request(command: CommandTypes.GetCodeFixes, args: protocol.CodeFixRequestArgs): Promise<protocol.GetCodeFixesResponse>;
    public request(command: CommandTypes.Geterr, args: protocol.GeterrRequestArgs): Promise<protocol.RequestCompletedEvent>;
    public request(command: CommandTypes.GeterrForProject, args: protocol.GeterrForProjectRequestArgs): Promise<protocol.RequestCompletedEvent>;
    public request(command: CommandTypes.Navto, args: protocol.NavtoRequestArgs): Promise<protocol.NavtoResponse>;
    public request(command: CommandTypes.NavTree, args: protocol.FileRequestArgs): Promise<protocol.NavTreeResponse>;
    public request(command: CommandTypes.Completions, args: protocol.CompletionsRequestArgs): Promise<protocol.CompletionsResponse>;
    public request(command: CommandTypes.CompletionDetails, args: protocol.CompletionDetailsRequestArgs): Promise<protocol.CompletionDetailsResponse>;
    public request(command: CommandTypes.DocumentHighlights, args: protocol.DocumentHighlightsRequestArgs): Promise<protocol.DocumentHighlightsResponse>;
    public request(command: CommandTypes.Quickinfo, args: protocol.FileLocationRequestArgs): Promise<protocol.QuickInfoResponse>;
    public request(command: CommandTypes.Rename, args: protocol.RenameRequestArgs): Promise<protocol.RenameResponse>;
    public request(command: CommandTypes.References, args: protocol.FileLocationRequestArgs): Promise<protocol.ReferencesResponse>;
    public request(command: CommandTypes.SignatureHelp, args: protocol.SignatureHelpRequestArgs): Promise<protocol.SignatureHelpResponse>;
    public request(command: string, args: object): Promise<object> {
        return this.sendMessage(command, false, args)!;
    }

    protected sendMessage(command: string, notification: boolean, args?: any): Promise<any> | undefined {
        this.seq = this.seq + 1;
        let request: protocol.Request = {
            command,
            seq: this.seq,
            type: 'request',
        };
        if (args) {
            request.arguments = args;
        }
        const serializedRequest = JSON.stringify(request) + '\n';
        console.log('serializedRequest', serializedRequest);
        this.tsserverProc.stdin.write(serializedRequest);
        if (notification) {
            return;
        } else {
            return (this.deferreds[this.seq] = new Deferred<any>(command)).promise;
        }
    }

    protected processMessage(untrimmedMessageString: string): void {
        const messageString = untrimmedMessageString.trim();
        if (!messageString || messageString.startsWith('Content-Length:')) {
            return;
        }
        const message: protocol.Message = JSON.parse(messageString);
        if (this.isResponse(message)) {
            this.resolveResponse(message, message.request_seq, message.success);
        } else if (this.isEvent(message)) {
            if (this.isRequestCompletedEvent(message)) {
                this.resolveResponse(message, message.body.request_seq, true);
            } else {
                // console.log('message', message);
                if (this.options.onEvent) {
                    this.options.onEvent(message);
                }
            }
        }
    }

    private resolveResponse(message: protocol.Message, request_seq: number, success: boolean) {
        const deferred = this.deferreds[request_seq];
        // console.log('has deferred', !!deferred, message, Object.keys(this.deferreds));
        if (deferred) {
            if (success) {
                this.deferreds[request_seq].resolve(message);
            } else {
                this.deferreds[request_seq].reject(message);
            }
            delete this.deferreds[request_seq];
        }
    }

    private isEvent(message: protocol.Message): message is protocol.Event {
        return message.type === 'event';
    }

    private isResponse(message: protocol.Message): message is protocol.Response {
        return message.type === 'response';
    }

    private isRequestCompletedEvent(message: protocol.Message): message is protocol.RequestCompletedEvent {
        return this.isEvent(message) && message.event === 'requestCompleted';
    }
}